@@ -10,7 +10,7 @@ import sys
 import time
 from datetime import datetime, timedelta
 from string import Template
-from typing import List, Optional
+from typing import List, Optional, Tuple
 
 from configparser import ConfigParser, ExtendedInterpolation
 from urllib.parse import urlparse
@@ -72,6 +72,7 @@ class HttpdTestSetup:
         self._source_dirs = [os.path.dirname(inspect.getfile(HttpdTestSetup))]
         self._modules = HttpdTestSetup.MODULES.copy()
         self._optional_modules = []
+        self._local_modules = []
 
     def add_source_dir(self, source_dir):
         self._source_dirs.append(source_dir)
@@ -82,6 +83,9 @@ class HttpdTestSetup:
     def add_optional_modules(self, modules: List[str]):
         self._optional_modules.extend(modules)
 
+    def add_local_module(self, mod_name: str, mod_path: str):
+        self._local_modules.append((mod_name, mod_path))
+
     def make(self):
         self._make_dirs()
         self._make_conf()
@@ -92,7 +96,9 @@ class HttpdTestSetup:
             self.add_modules([self.env.ssl_module])
         self._make_modules_conf()
         self._make_htdocs()
+        self._add_local_modules()
         self._add_aptest()
+        self._build_clients()
         self.env.clear_curl_headerfiles()
 
     def _make_dirs(self):
@@ -196,6 +202,24 @@ class HttpdTestSetup:
             # load our test module which is not installed
             fd.write(f"LoadModule aptest_module   \"{local_dir}/mod_aptest/.libs/mod_aptest.so\"\n")
 
+    def _add_local_modules(self):
+        if len(self._local_modules):
+            modules_conf = os.path.join(self.env.server_dir, 'conf/modules.conf')
+            with open(modules_conf, 'a') as fd:
+                for mod_name, mod_path in self._local_modules:
+                    mod_path = os.path.join(self.env.src_dir, mod_path)
+                    fd.write(f"LoadModule {mod_name}_module   \"{mod_path}\"\n")
+
+    def _build_clients(self):
+        clients_dir = os.path.join(
+            os.path.dirname(os.path.dirname(inspect.getfile(HttpdTestSetup))),
+            'clients')
+        p = subprocess.run(['make'], capture_output=True, cwd=clients_dir)
+        rv = p.returncode
+        if rv != 0:
+            log.error(f"compiling test clients failed: {p.stderr}")
+            raise Exception(f"compiling test clients failed: {p.stderr}")
+
 
 class HttpdTestEnv:
 
@@ -233,7 +257,7 @@ class HttpdTestEnv:
         self._bin_dir = self.config.get('global', 'bindir')
         self._apxs = self.config.get('global', 'apxs')
         self._prefix = self.config.get('global', 'prefix')
-        self._apachectl = self.config.get('global', 'apachectl')
+        self._httpd = self.config.get('global', 'httpd')
         if HttpdTestEnv.LIBEXEC_DIR is None:
             HttpdTestEnv.LIBEXEC_DIR = self._libexec_dir = self.get_apxs_var('LIBEXECDIR')
         self._curl = self.config.get('global', 'curl_bin')
@@ -250,8 +274,11 @@ class HttpdTestEnv:
         self._http_port2 = int(self.config.get('test', 'http_port2'))
         self._https_port = int(self.config.get('test', 'https_port'))
         self._proxy_port = int(self.config.get('test', 'proxy_port'))
+        self._ws_port = int(self.config.get('test', 'ws_port'))
         self._http_tld = self.config.get('test', 'http_tld')
         self._test_dir = self.config.get('test', 'test_dir')
+        self._clients_dir = os.path.join(os.path.dirname(self._test_dir), 'clients')
+        self._src_dir = self.config.get('test', 'src_dir')
         self._gen_dir = self.config.get('test', 'gen_dir')
         self._server_dir = os.path.join(self._gen_dir, 'apache')
         self._server_conf_dir = os.path.join(self._server_dir, "conf")
@@ -259,7 +286,7 @@ class HttpdTestEnv:
         self._server_logs_dir = os.path.join(self.server_dir, "logs")
         self._server_access_log = os.path.join(self._server_logs_dir, "access_log")
         self._error_log = HttpdErrorLog(os.path.join(self._server_logs_dir, "error_log"))
-        self._apachectl_stderr = None
+        self._httpd_cmd_stderr = None
 
         self._dso_modules = self.config.get('httpd', 'dso_modules').split(' ')
         self._mpm_modules = self.config.get('httpd', 'mpm_modules').split(' ')
@@ -322,6 +349,12 @@ class HttpdTestEnv:
             for name in self._httpd_log_modules:
                 self._log_interesting += f" {name}:{log_level}"
 
+    def check_error_log(self):
+        errors, warnings = self._error_log.get_missed()
+        assert (len(errors), len(warnings)) == (0, 0),\
+                f"apache logged {len(errors)} errors and {len(warnings)} warnings: \n"\
+                "{0}\n{1}\n".format("\n".join(errors), "\n".join(warnings))
+
     @property
     def curl(self) -> str:
         return self._curl
@@ -366,6 +399,10 @@ class HttpdTestEnv:
     def proxy_port(self) -> int:
         return self._proxy_port
 
+    @property
+    def ws_port(self) -> int:
+        return self._ws_port
+
     @property
     def http_tld(self) -> str:
         return self._http_tld
@@ -386,10 +423,18 @@ class HttpdTestEnv:
     def gen_dir(self) -> str:
         return self._gen_dir
 
+    @property
+    def src_dir(self) -> str:
+        return self._src_dir
+
     @property
     def test_dir(self) -> str:
         return self._test_dir
 
+    @property
+    def clients_dir(self) -> str:
+        return self._clients_dir
+
     @property
     def server_dir(self) -> str:
         return self._server_dir
@@ -442,7 +487,7 @@ class HttpdTestEnv:
 
     @property
     def apachectl_stderr(self):
-        return self._apachectl_stderr
+        return self._httpd_cmd_stderr
 
     def add_cert_specs(self, specs: List[CertificateSpec]):
         self._cert_specs.extend(specs)
@@ -519,12 +564,17 @@ class HttpdTestEnv:
         if not os.path.exists(path):
             return os.makedirs(path)
 
-    def run(self, args, stdout_list=False, intext=None, debug_log=True):
+    def run(self, args, stdout_list=False, intext=None, inbytes=None, debug_log=True,
+            run_env=None):
+        if not run_env:
+            run_env = os.environ
         if debug_log:
             log.debug(f"run: {args}")
         start = datetime.now()
+        if intext is not None:
+            inbytes = intext.encode()
         p = subprocess.run(args, stderr=subprocess.PIPE, stdout=subprocess.PIPE,
-                           input=intext.encode() if intext else None)
+                           input=inbytes, env=run_env)
         stdout_as_list = None
         if stdout_list:
             try:
@@ -549,8 +599,18 @@ class HttpdTestEnv:
         with open(self._test_conf, 'w') as fd:
             fd.write('\n'.join(self._httpd_base_conf))
             fd.write('\n')
-            if self._verbosity >= 2:
-                fd.write(f"LogLevel core:trace5 {self.mpm_module}:trace5 http:trace5\n")
+            fd.write(f"CoreDumpDirectory {self._server_dir}\n")
+            fd.write('\n')
+            if self._verbosity >= 3:
+                fd.write(f"LogLevel trace7 ssl:trace6\n")
+                fd.write(f"DumpIoOutput on\n")
+                fd.write(f"DumpIoInput on\n")
+            elif self._verbosity >= 2:
+                fd.write(f"LogLevel debug core:trace5 {self.mpm_module}:trace5 ssl:trace5 http:trace5\n")
+            elif self._verbosity >= 1:
+                fd.write(f"LogLevel info\n")
+            else:
+                fd.write(f"LogLevel warn\n")
             if self._log_interesting:
                 fd.write(self._log_interesting)
             fd.write('\n\n')
@@ -567,7 +627,7 @@ class HttpdTestEnv:
         while datetime.now() < try_until:
             # noinspection PyBroadException
             try:
-                r = self.curl_get(url, insecure=True)
+                r = self.curl_get(url, insecure=True, options=['-vvvv'])
                 if r.exit_code == 0:
                     return True
                 time.sleep(.1)
@@ -607,49 +667,68 @@ class HttpdTestEnv:
         log.debug(f"Server still responding after {timeout}")
         return False
 
-    def _run_apachectl(self, cmd) -> ExecResult:
+    def _httpd_cmd(self, cmd) -> ExecResult:
         conf_file = 'stop.conf' if cmd == 'stop' else 'httpd.conf'
-        args = [self._apachectl,
+        env = os.environ.copy()
+        args = [self._httpd,
                 "-d", self.server_dir,
                 "-f", os.path.join(self._server_dir, f'conf/{conf_file}'),
                 "-k", cmd]
-        r = self.run(args)
-        self._apachectl_stderr = r.stderr
+        r = self.run(args, run_env=env)
+        self._httpd_cmd_stderr = r.stderr
         if r.exit_code != 0:
             log.warning(f"failed: {r}")
         return r
 
     def apache_reload(self):
-        r = self._run_apachectl("graceful")
+        r = self._httpd_cmd("graceful")
         if r.exit_code == 0:
             timeout = timedelta(seconds=10)
-            return 0 if self.is_live(self._http_base, timeout=timeout) else -1
+            if self.is_live(self._http_base, timeout=timeout):
+                return 0
+            log.error('failed to reload apache')
+            self._error_log.dump(log)
+            return -1
         return r.exit_code
 
     def apache_restart(self):
-        self.apache_stop()
-        r = self._run_apachectl("start")
+        x = self.apache_stop()
+        if x != 0:
+            return x
+        r = self._httpd_cmd("start")
         if r.exit_code == 0:
             timeout = timedelta(seconds=10)
-            return 0 if self.is_live(self._http_base, timeout=timeout) else -1
+            if self.is_live(self._http_base, timeout=timeout):
+                return 0
+            log.error('failed to reload apache')
+            self._error_log.dump(log)
+            return -1
         return r.exit_code
         
     def apache_stop(self):
-        r = self._run_apachectl("stop")
+        r = self._httpd_cmd("stop")
         if r.exit_code == 0:
             timeout = timedelta(seconds=10)
-            return 0 if self.is_dead(self._http_base, timeout=timeout) else -1
+            if self.is_dead(self._http_base, timeout=timeout):
+                return 0
+            log.error('failed to stop apache')
+            self._error_log.dump(log)
+            return -1
         return r
 
     def apache_graceful_stop(self):
         log.debug("stop apache")
-        self._run_apachectl("graceful-stop")
-        return 0 if self.is_dead() else -1
+        self._httpd_cmd("graceful-stop")
+        if self.is_dead():
+            return 0
+        log.error('failed to gracefully stop apache')
+        self._error_log.dump(log)
+        return -1
 
     def apache_fail(self):
         log.debug("expect apache fail")
-        self._run_apachectl("stop")
-        rv = self._run_apachectl("start")
+        self._httpd_cmd("stop")
+        rv = self._httpd_cmd("start")
         if rv == 0:
             rv = 0 if self.is_dead() else -1
         else:
@@ -690,7 +769,7 @@ class HttpdTestEnv:
         if force_resolve and u.hostname and u.hostname != 'localhost' \
                 and u.hostname != self._httpd_addr \
                 and not re.match(r'^(\d+|\[|:).*', u.hostname):
-            assert u.port, f"port not in url: {urls[0]}"
+            assert u.port, f"port not in url: {url}"
             args.extend(["--resolve", f"{u.hostname}:{u.port}:{self._httpd_addr}"])
         return args
 
